import torch
from torch.nn import functional as F
from torch.distributions import Normal, kl_divergence as kldiv
import torch
import torch.nn as nn
import torch.nn.functional as F
from sklearn.metrics.pairwise import rbf_kernel
from torch.autograd import Variable

from ._logger import *
from ._distributions import *
from ._compat import *
from ..externals._trvae_mmd_loss import mmd_loss_calc

# Reference: https://github.com/tim-learn/ATDOC/blob/main/loss.py
class MMD(nn.Module):
    def __init__(self, kernel_type='rbf', kernel_mul=2.0, kernel_num=5):
        super(MMD, self).__init__()
        self.kernel_num = kernel_num
        self.kernel_mul = kernel_mul
        self.fix_sigma = None
        self.kernel_type = kernel_type

    def guassian_kernel(self, source, target, kernel_mul=2.0, kernel_num=5, fix_sigma=None):
        n_samples = int(source.size()[0]) + int(target.size()[0])
        total = torch.cat([source, target], dim=0)
        total0 = total.unsqueeze(0).expand(int(total.size(0)), int(total.size(0)), int(total.size(1)))
        total1 = total.unsqueeze(1).expand(int(total.size(0)), int(total.size(0)), int(total.size(1)))
        L2_distance = ((total0-total1)**2).sum(2)
        if fix_sigma:
            bandwidth = fix_sigma
        else:
            bandwidth = torch.sum(L2_distance.data) / (n_samples**2-n_samples)
        bandwidth /= kernel_mul ** (kernel_num // 2)
        bandwidth_list = [bandwidth * (kernel_mul**i) for i in range(kernel_num)]
        kernel_val = [torch.exp(-L2_distance / bandwidth_temp) for bandwidth_temp in bandwidth_list]
        return sum(kernel_val)

    def linear_mmd2(self, f_of_X, f_of_Y):
        loss = 0.0
        delta = f_of_X - f_of_Y
        loss = torch.mean((delta[:-1] * delta[1:]).sum(1))
        return loss

    def forward(self, source, target):
        if self.kernel_type == 'linear':
            return self.linear_mmd2(source, target)
        elif self.kernel_type == 'rbf':
            batch_size = int(source.size()[0])
            kernels = self.guassian_kernel(
                source, target, kernel_mul=self.kernel_mul, kernel_num=self.kernel_num, fix_sigma=self.fix_sigma)
            with torch.no_grad():
                XX = torch.mean(kernels[:batch_size, :batch_size])
                YY = torch.mean(kernels[batch_size:, batch_size:])
                XY = torch.mean(kernels[:batch_size, batch_size:])
                YX = torch.mean(kernels[batch_size:, :batch_size])
                loss = torch.mean(XX + YY - XY - YX)
            torch.cuda.empty_cache()
            return loss

class LossFunction:
    @staticmethod
    def mse(recon_x:   torch.tensor, 
            x:         torch.tensor, 
            reduction: str = "sum"):
        """
        The reconstruction error in the form of mse
        """
        return F.mse_loss(recon_x, x, reduction = reduction)

    @staticmethod
    def bce(recon_x:   torch.tensor, 
            x:         torch.tensor, 
            reduction: str = "sum"):
        """
        The reconstruction error in the form of mse
        """
        return F.binary_cross_entropy(recon_x, x)

    @staticmethod
    def vae_mse(recon_x: torch.tensor, 
                x:       torch.tensor, 
                mu:      torch.tensor, 
                var:     torch.tensor, 
                kld_weight: float = 0.5):
        """
        The KL-divergence of the latent probability z
        KL(q || p) = -∫ q(z) log [ p(z) / q(z) ] 
                = -E[log p(z) - log q(z)] 
        """
        MSE = F.mse_loss(recon_x, x, reduction = "sum")
        KLD = torch.mean(-0.5 * torch.sum(1 + var - mu ** 2 - var.exp(), dim = 1), dim = 0)

        return MSE + KLD
        
    @staticmethod
    def mmd_loss(x: torch.tensor, y: torch.tensor, dim = -1):
        if dim == -1:
            return MMD()(x,y)
        elif dim == 1:
            result = torch.tensor(0.)
            for i in range(x.shape[1]):
                result += MMD()(x[:,i],y[:,i])
            return result
        else:
            raise ValueError("dim must be -1 or 1")
    
    @staticmethod 
    def mmd_loss_trvae(x: torch.tensor, y: torch.tensor):
        return mmd_loss_calc(x,y)
    
    @staticmethod
    def zinb_reconstruction_loss(X:            torch.tensor, 
                                 total_counts: torch.tensor = None,
                                 logits:       torch.tensor = None,
                                 mu:           torch.tensor = None,
                                 theta:        torch.tensor = None,
                                 gate_logits:  torch.tensor = None,
                                 reduction:    str = "sum"):
        if ((total_counts == None) and (logits == None)):
            if ((mu == None) and (theta == None )):
                raise ValueError
            logits = (mu / theta).log()
            total_counts = theta + 1e-6
            znb = ZeroInflatedNegativeBinomial(
                total_count=total_counts, 
                logits=logits,
                gate_logits=gate_logits
            )   
        else: 
            znb = ZeroInflatedNegativeBinomial(
                total_count=total_counts, 
                logits=logits, 
                gate_logits=gate_logits
            )
        if reduction == "sum":
            reconst_loss = -znb.log_prob(X).sum(dim = 1)
        elif reduction == "mean":
            reconst_loss = -znb.log_prob(X).mean(dim = 1)
        elif reduction == "none":
            reconst_loss = -znb.log_prob(X)
        return reconst_loss

    @staticmethod
    def nb_reconstruction_loss(X:            torch.tensor, 
                               total_counts: torch.tensor = None,
                               logits:       torch.tensor = None,
                               mu:           torch.tensor = None,
                               theta:        torch.tensor = None,
                               reduction:    str = "sum"):
        if ((total_counts == None) and (logits == None)):
            if ((mu == None) and (theta == None )):
                raise ValueError
            logits = (mu + 1e-6) - (theta + 1e-6).log()
            total_counts = theta 
        
        nb = NegativeBinomial(
            total_count=total_counts, 
            logits=logits, 
        )
        if reduction == "sum":
            reconst_loss = -nb.log_prob(X).sum(dim = 1)
        elif reduction == "mean":
            reconst_loss = -nb.log_prob(X).mean(dim = 1)
        elif reduction == "none":
            reconst_loss = -nb.log_prob(X)
        return reconst_loss

    @staticmethod 
    def zi_gaussian_reconstruction_loss(
        X,
        mean,
        variance,
        gate_logits,
        reduction: Literal['sum','mean'] = 'sum'
    ):
        zg = ZeroInflatedGaussian(
            mean=mean,
            variance=variance,
            gate_logits=gate_logits
        )   
        if reduction == "sum":
            reconst_loss = -zg.log_prob(X).sum(dim = 1)
        elif reduction == "mean":
            reconst_loss = -zg.log_prob(X).mean(dim = 1)
        return reconst_loss 

    @staticmethod
    def kld(q: torch.tensor,
            p: torch.tensor):
        kl_loss = kldiv(q.log(), p, reduction="sum", log_target=False)
        kl_loss.requires_grad_(True)
        return kl_loss

    @staticmethod
    def kl1(mu:  torch.tensor, 
            var: torch.tensor):
        return kldiv(Normal(mu, torch.sqrt(var)), Normal(0, 1)).sum(dim=1)

    @staticmethod
    def kl2(mu1: torch.tensor, 
           var1: torch.tensor, 
           mu2:  torch.tensor, 
           var2: torch.tensor):
        return kldiv(Normal(mu1, var1.sqrt()), Normal(mu2, var2.sqrt()))

    @staticmethod
    def soft_cross_entropy(pred:         torch.tensor, 
                           soft_targets: torch.tensor):
        # use nn.CrossEntropyLoss if not using soft labels
        logsoftmax = nn.LogSoftmax(dim=1)
        return torch.mean(torch.sum(- soft_targets * logsoftmax(pred), 1))



import atexit
import time
from functools import reduce
import tqdm

verbosity = 1
""" Global verbosity level, choose from {0,...,6}. """

class Colors:
    BLACK   = "\033[0;30m"
    DGRAY   = "\033[1;30m"
    RED     = "\033[0;31m"
    LRED    = "\033[1;31m"
    GREEN   = "\033[0;32m"
    LGREEN  = "\033[1;32m"
    ORANGE  = "\033[0;33m"
    YELLOW  = "\033[1;33m"
    BLUE    = "\033[0;34m"
    LBLUE   = "\033[1;34m"
    PURPLE  = "\033[0;35m"
    LPURPLE = "\033[1;35m"
    CYAN    = "\033[0;36m"
    LCYAN   = "\033[1;36m"
    LGRAY   = "\033[0;37m"
    WHITE   = "\033[1;37m"
    BRED    = "\033[0;37;41m"
    BGREEN  = "\033[0;37;42m"
    BYELLOW = "\033[0;37;43m"
    BBLUE   = "\033[0;37;44m"
    NC      = "\033[0m"



def filename(filename=''):
    """ 
    @brief Define filename of logfile. 
           If not defined, log output will be to the standard output.

    @arg filename : str
    """
    global logfilename
    logfilename = filename
    # if providing a logfile name, automatically set verbosity to a very high level
    verbosity(5)

def m(v=0,*msg):
    """ 
    @brief Write message to log output, depending on verbosity level.

    @arg v : int
        Verbosity level of message.
    @arg *msg : 
        One or more arguments to be formatted as string. Same behavior as print
        function.
    """
    if verbosity > v:
        print(time.asctime(), end = '\t')
        mi(*msg)


def mi(*msg):
    """ 
    @brief Write message to log output, ignoring the verbosity level.

    @arg *msg : 
        One or more arguments to be formatted as string. Same behavior as print
        function.
    """
    if logfilename == '':
        # in python 3, the following works
        # print(*msg)
        # due to compatibility with the print statement in python 2 we choose
        print(' '.join([str(m) for m in msg]))
    else:
        out = ''
        for s in msg:
            out += str(s) + ' '
        with file(filename, 'a') as f:
            f.write(out + '\n')

def mt(*msg):
    """ 
    @brief Write message to log output, ignoring the verbosity level.

    @brief *msg : 
        One or more arguments to be formatted as string. Same behavior as print
        function.
    """
    if logfilename == '':
        # in python 3, the following works
        # print(*msg)
        # due to compatibility with the print statement in python 2 we choose
        print(time.asctime() + ' - TDI - INFO - ', end = '\t')
        print(' '.join([str(m) for m in msg]))
    else:
        out = ''
        for s in msg:
            out += str(s) + ' '
        with file(filename, 'a') as f:
            f.write(time.asctime() + '\t')
            f.write(out + '\n')

def mw(*msg):
    """ 
    @brief Write message to log output, ignoring the verbosity level.

    @brief *msg : 
        One or more arguments to be formatted as string. Same behavior as print
        function.
    """
    if logfilename == '':
        # in python 3, the following works
        # print(*msg)
        # due to compatibility with the print statement in python 2 we choose
        print(time.asctime() + ' - TDI - WARNING - ', end = '\t')
        print(Colors.YELLOW + ' '.join([str(m) for m in msg]) + Colors.NC)
    else:
        out = ''
        for s in msg:
            out += str(s) + ' '
        with file(filename, 'a') as f:
            f.write(time.asctime() + '\t')
            f.write(out + '\n')

def _sec_to_str(t):
    """ 
    @brief Format time in seconds.

    @arg t : int
        Time in seconds.
    """
    return "%d:%02d:%02d.%03d" % \
        reduce(lambda ll,b : divmod(ll[0],b) + ll[1:],
            [(t*1000,),1000,60,60])


def is_notebook() -> bool:
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter

def get_tqdm():
    if is_notebook():
        from tqdm.autonotebook import tqdm as tqdm_notebook
        return tqdm_notebook
    else:
        return tqdm.tqdm

def _terminate():
    """ 
    @brief  Function called when program terminates.
            Similar to mt, but writes total runtime.
    """
    if verbosity > 0 and not is_notebook():
        now = time.time()
        elapsed_since_start = now - start
        # mt(_sec_to_str(elapsed_since_start),'- total runtime')


# further global variables
start = time.time() # time.time() is deprecated since version python version 3.3
intermediate = start
logfilename = ''
separator = 40*"-"

atexit.register(_terminate)



import os
from pathlib import Path
import scanpy as sc
import warnings
MODULE_PATH = Path(__file__).parent
warnings.filterwarnings("ignore")

zenodo_accession = '8196645'


################
#     Human    #
################


def human_gex_reference_v2():
    """
    Load the human gex reference v2
    """
    default_path = os.path.join(MODULE_PATH, '../data/datasets/human_gex_reference_v2.h5ad')
    if os.path.exists(
        default_path
    ):
        return sc.read_h5ad(default_path)
    else:
        os.system(f"curl -o {default_path} https://zenodo.org/record/{zenodo_accession}/files/human_gex_reference_v2.h5ad?download=1")
    
def human_gex_reference_v2_cd4():
    """
    Load the human gex reference v2 CD4 subset
    """
    default_path = os.path.join(MODULE_PATH, '../data/datasets/human_gex_reference_v2_cd4.h5ad')
    if os.path.exists(
        default_path
    ):
        return sc.read_h5ad(default_path)
    else:
        os.system(f"curl -o {default_path} https://zenodo.org/record/{zenodo_accession}/files/human_gex_reference_v2_cd4.h5ad?download=1")

def human_gex_reference_v2_cd8():
    """
    Load the human gex reference v2 CD8 subset
    """
    default_path = os.path.join(MODULE_PATH, '../data/datasets/human_gex_reference_v2_cd8.h5ad')
    if os.path.exists(
        default_path
    ):
        return sc.read_h5ad(default_path)
    else:
        os.system(f"curl -o {default_path} https://zenodo.org/record/{zenodo_accession}/files/human_gex_reference_v2_cd8.h5ad?download=1")


def human_gex_reference_v2_zheng_2020_cd4():
    """
    Load the human gex reference v2 merged with Zheng et al. 2020 CD4 subset
    """
    default_path = os.path.join(MODULE_PATH, '../data/datasets/human_gex_reference_v2_zheng_2020_cd4.h5ad')
    if os.path.exists(
        default_path
    ):
        return sc.read_h5ad(default_path)
    else:
        os.system(f"curl -o {default_path} https://zenodo.org/record/{zenodo_accession}/files/human_gex_reference_v2_zheng_2020_cd4.h5ad?download=1")

def human_gex_reference_v2_zheng_2020_cd8():
    """
    Load the human gex reference v2 merged with Zheng et al. 2020 CD8 subset
    """
    default_path = os.path.join(MODULE_PATH, '../data/datasets/human_gex_reference_v2_zheng_2020_cd8.h5ad')
    if os.path.exists(
        default_path
    ):
        return sc.read_h5ad(default_path)
    else:
        os.system(f"curl -o {default_path} https://zenodo.org/record/{zenodo_accession}/files/human_gex_reference_v2_zheng_2020_cd8.h5ad?download=1")

def human_gex_multi_atlas_v1_cd4():
    """
    Load the human gex multi atlas v1 CD4 subset
    """
    default_path = os.path.join(MODULE_PATH, '../data/datasets/human_gex_multi_atlas_v1_cd4.h5ad')
    if os.path.exists(
        default_path
    ):
        return sc.read_h5ad(default_path)
    else:
        raise FileNotFoundError("Dataset has not been published online. Please contact the author for more information.")
    
def human_gex_multi_atlas_v1_cd8():
    """
    Load the human gex multi atlas v1 CD8 subset
    """
    default_path = os.path.join(MODULE_PATH, '../data/datasets/human_gex_multi_atlas_v1_cd8.h5ad')
    if os.path.exists(
        default_path
    ):
        return sc.read_h5ad(default_path)
    else:
        raise FileNotFoundError("Dataset has not been published online. Please contact the author for more information.")


################
#     Model    #
################


def download_model_weights():
    os.system(f"curl -o ../data/pretrained_weights/human_bert_tcr_768_v2.ckpt https://zenodo.org/record/{zenodo_accession}/files/human_bert_tcr_768_v2.ckpt?download=1")
    os.system(f"curl -o ../data/pretrained_weights/human_bert_tcr_pca_v2.pkl https://zenodo.org/record/{zenodo_accession}/files/human_bert_tcr_pca_v2.pkl?download=1")




import torch
from torch.autograd import Variable

def pairwise_distance(x, y):
    x = x.view(x.shape[0], x.shape[1], 1)
    y = torch.transpose(y, 0, 1)
    output = torch.sum((x - y) ** 2, 1)
    output = torch.transpose(output, 0, 1)

    return output


def gaussian_kernel_matrix(x, y, alphas):
    """Computes multiscale-RBF kernel between x and y.

       Parameters
       ----------
       x: torch.Tensor
            Tensor with shape [batch_size, z_dim].
       y: torch.Tensor
            Tensor with shape [batch_size, z_dim].
       alphas: Tensor

       Returns
       -------
       Returns the computed multiscale-RBF kernel between x and y.
    """

    dist = pairwise_distance(x, y).contiguous()
    dist_ = dist.view(1, -1)

    alphas = alphas.view(alphas.shape[0], 1)
    beta = 1. / (2. * alphas)

    s = torch.matmul(beta, dist_)

    return torch.sum(torch.exp(-s), 0).view_as(dist)


def mmd_loss_calc(source_features, target_features):
    """Initializes Maximum Mean Discrepancy(MMD) between source_features and target_features.

       - Gretton, Arthur, et al. "A Kernel Two-Sample Test". 2012.

       Parameters
       ----------
       source_features: torch.Tensor
            Tensor with shape [batch_size, z_dim]
       target_features: torch.Tensor
            Tensor with shape [batch_size, z_dim]

       Returns
       -------
       Returns the computed MMD between x and y.
    """
    alphas = [
        1e-6, 1e-5, 1e-4, 1e-3, 1e-2, 1e-1, 1, 5, 10, 15, 20, 25, 30, 35, 100,
        1e3, 1e4, 1e5, 1e6
    ]
    alphas = Variable(torch.FloatTensor(alphas)).to(device=source_features.device)

    cost = torch.mean(gaussian_kernel_matrix(source_features, source_features, alphas))
    cost += torch.mean(gaussian_kernel_matrix(target_features, target_features, alphas))
    cost -= 2 * torch.mean(gaussian_kernel_matrix(source_features, target_features, alphas))

    return cost


from ._metadata import __version__, within_flit
import importlib
import subprocess
import warnings


if not within_flit():  # see function docstring on why this is there
    # the actual API
    # (start with settings as several tools are using it)
    from . import model as model
    from . import preprocessing as pp
    from . import tools as tl
    from . import data as data
    from . import utils as ut
    from . import pipeline as pipeline


    from anndata import AnnData, concat
    from anndata import (
        read_h5ad,
        read_csv,
        read_excel,
        read_hdf,
        read_loom,
        read_mtx,
        read_text,
        read_umi_tools,
    )


    # has to be done at the end, after everything has been imported
    import sys
    sys.modules.update({f'{__name__}.{m}': globals()[m] for m in ['model', 'pp', 'tl', 'data', 'ut', 'pipeline']})
    del sys


import os
import argparse
import sys
import numpy as np
import random
import pandas as pd
from typing import Iterable
import torch
import torch.nn.functional as F
from collections import Counter
import scanpy as sc
import numba
from umap.distances import euclidean
import scipy 
import math 
import tabulate
import warnings
import tqdm
from ._logger import Colors
from ._parallelizer import Parallelizer
from ._decorators import deprecated

is_nvidia_smi_warned = False

def convert_size(size_bytes):
   if size_bytes == 0:
       return "0B"
   size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
   i = int(math.floor(math.log(size_bytes, 1024)))
   p = math.pow(1024, i)
   s = round(size_bytes / p, 2)
   return "%s %s" % (s, size_name[i])

def print_gpu_mem(i):
    global is_nvidia_smi_warned
    try:
        import nvidia_smi
    except:
        if not is_nvidia_smi_warned:
            print("install nvidia_smi with pip install nvidia-smi for automatically select cuda device by memory usage.")
            is_nvidia_smi_warned = True
        return "0", "0", "0 %"
    nvidia_smi.nvmlInit()
    handle = nvidia_smi.nvmlDeviceGetHandleByIndex(i)
    info = nvidia_smi.nvmlDeviceGetMemoryInfo(handle)
    return convert_size(info.used), convert_size(info.total), '%.2f' % ((info.used / info.total) * 100) + ' %'


def get_default_device():
    """
    Get the default device for training
    """
    devices = {}
    n_cuda_device = torch.cuda.device_count()
    if n_cuda_device > 0:
        for i in range(n_cuda_device):
            used, total, perc = print_gpu_mem(i)
            devices['cuda:' + str(i)] = {
                'name': torch.cuda.get_device_name(i),
                'used': used,
                'total': total,
                'used %': perc
            }
    else:
        return 'cpu'
    selected_device = sorted(devices.items(), key=lambda x: float(x[1]['used %'].split(" ")[0]))[0][0]
    for k,v in devices.items():
        if k == selected_device:
            v['selected'] = '*'
        else:
            v['selected'] = ' '
    print(tabulate.tabulate(pd.DataFrame(devices).T))
    return selected_device


def FLATTEN(x): 
    return [i for s in x for i in s]


def multi_values_dict(keys, values):
    ret = {}
    for k,v in zip(keys, values):
        if k not in ret.keys():
            ret[k] = [v]
        else:
            ret[k].append(v)
    return ret


def random_subset_by_key_fast(adata, key, n):
    from collections import Counter
    counts = {k:v/len(adata) for k,v in Counter(adata.obs[key]).items()}
    ns = [(k,int(v*n)) for k,v in counts.items()]
    all_indices = []
    for k,v in ns:
        indices = np.argwhere(np.array(adata.obs[key] == k)).flatten()
        if len(indices) > 0:
            indices = np.random.choice(indices, v, replace=False)
            all_indices.append(indices)
    all_indices = np.hstack(all_indices)
    return adata[all_indices]

def exists(x):
    return x != None


def print_version():
    print(Colors.YELLOW)
    print('Python VERSION:{}\n'.format(Colors.NC), sys.version)
    print(Colors.YELLOW)
    print('PyTorch VERSION:{}\n'.format(Colors.NC), torch.__version__)
    print(Colors.YELLOW)
    print('CUDA VERSION{}\n'.format(Colors.NC))
    from subprocess import call
    try: call(["nvcc", "--version"])
    except: pass
    print(Colors.YELLOW)
    print('CUDNN VERSION:{}\n'.format(Colors.NC), torch.backends.cudnn.version())
    print(Colors.YELLOW)
    print('Number CUDA Devices:{}\n'.format(Colors.NC), torch.cuda.device_count())
    try:
        print('Devices             ')
        call(["nvidia-smi", "--format=csv", "--query-gpu=index,name,driver_version,memory.total,memory.used,memory.free"])
    except FileNotFoundError:
        # There is no nvidia-smi in this machine
        pass
    if torch.cuda.is_available():
        print('Active CUDA Device: GPU', torch.cuda.current_device())
        print ('Available devices     ', torch.cuda.device_count())
        print ('Current cuda device   ', torch.cuda.current_device())
    else:
        # cuda not available
        pass

def readMM(path):
    f = open(path)
    while 1:
        line = f.readline()
        if not line.startswith("%"):
            break 

    n,m,total=line.strip().split(" ")
    n = int(n)
    m = int(m)
    total = int(total)
    z = np.zeros((n,m),dtype=np.int32)
    pbar = tqdm.tqdm(total=total)
    while 1:
        line = f.readline()
        if not line:
            break
        a,b,c = line.split(' ')
        a = int(a)
        b = int(b)
        c = int(c)
        z[a-1,b-1]=c
        pbar.update(1)
    pbar.close()
    return scipy.sparse.csr_matrix(z.T)

def iter_spmatrix(matrix):
    """ Iterator for iterating the elements in a ``scipy.sparse.*_matrix`` 

    This will always return:
    >>> (row, column, matrix-element)

    Currently this can iterate `coo`, `csc`, `lil` and `csr`, others may easily be added.

    Parameters
    ----------
    matrix : ``scipy.sparse.sp_matrix``
      the sparse matrix to iterate non-zero elements
    """
    from scipy.sparse import isspmatrix_coo, isspmatrix_csc, isspmatrix_csr, isspmatrix_lil
    if isspmatrix_coo(matrix):
        for r, c, m in zip(matrix.row, matrix.col, matrix.data):
            yield r, c, m

    elif isspmatrix_csc(matrix):
        for c in range(matrix.shape[1]):
            for ind in range(matrix.indptr[c], matrix.indptr[c+1]):
                yield matrix.indices[ind], c, matrix.data[ind]

    elif isspmatrix_csr(matrix):
        for r in range(matrix.shape[0]):
            for ind in range(matrix.indptr[r], matrix.indptr[r+1]):
                yield r, matrix.indices[ind], matrix.data[ind]

    elif isspmatrix_lil(matrix):
        for r in range(matrix.shape[0]):
            for c, d in zip(matrix.rows[r], matrix.data[r]):
                yield r, c, d

    else:
        raise NotImplementedError("The iterator for this sparse matrix has not been implemented")

def writeMM(mat, path):
    with open(path, 'w') as f:
        f.write("%%MatrixMarket matrix coordinate integer general\n")
        f.write("% Generated by Snowxue\n")
        f.write("% \n")
        f.write(str(mat.shape[1]) + " " + str(mat.shape[0]) + " " + str(int(mat.nnz)) + "\n")
        pbar = tqdm.tqdm(total=mat.nnz,bar_format='{l_bar}{bar:10}{r_bar}{bar:-10b}',desc="writing matrix")
        for i,j,v in iter_spmatrix(mat):
            f.write(str(j+1) + " " + str(i+1) + " " + str(v) + "\n")
            pbar.update(1)
        pbar.close()



import os
import argparse
import sys
import numpy as np
import random
import pandas as pd
from typing import Iterable
import torch
import torch.nn.functional as F
from collections import Counter

def get_k_elements(arr: Iterable, k:int):
    return list(map(lambda x: x[k], arr))

def get_last_k_elements(arr: Iterable, k:int):
    return list(map(lambda x: x[k:], arr))

def get_elements(arr: Iterable, a:int, b:int):
    return list(map(lambda x: x[a:a+b], arr))

def one_hot_(labels, return_dict = False):
    n_labels_ = np.unique(labels)
    n_labels = dict(zip(n_labels_, range(len(n_labels_))))
    if return_dict:
        return {"one_hot": F.one_hot( torch.tensor(list(map(lambda x: n_labels[x], labels)))), "labels": n_labels}
    return F.one_hot( torch.tensor(list(map(lambda x: n_labels[x], labels))))

def one_hot(index: torch.Tensor, n_cat: int) -> torch.Tensor:
    """One hot a tensor of categories."""
    onehot = torch.zeros(index.size(0), n_cat, device=index.device)
    onehot.scatter_(1, index.type(torch.long), 1)
    return onehot.type(torch.float32)

    


from typing import Tuple
import functools 
import warnings
import inspect


def deprecated(*, ymd: Tuple[int] = None, optional_message: str = None):
    def decorate(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            warnings.simplefilter('always', DeprecationWarning)  # turn off filter
            warnings.warn("Function {} is going to be deprecated".format(func.__name__) + " after %d-%d-%d." % ymd if ymd else '.',
                        category=DeprecationWarning,
                        stacklevel = 2)
            warnings.simplefilter('default', DeprecationWarning)  # reset filter
            return func(*args, **kwargs)
        return wrapper 
    return decorate
        
def typed(types: dict = None, *, optional_message: str = None):
    def decorate(func):
        arg_names = inspect.getfullargspec(func)[0]
        @functools.wraps(func)
        def wrapper(*args,**kwargs):
            if types:
                for name, arg in zip(arg_names, args):
                    if name in types and not isinstance(arg, types[name]):
                        raise TypeError("Argument {} must be {}".format(name, types[name]))
                for name, arg in kwargs.items():
                    if name in types and not isinstance(arg, types[name]):
                        raise TypeError("Argument {} must be {}".format(name, types[name]))
            return func(*args, **kwargs)
        return wrapper
    return decorate
                
                
def timed(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        import time
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        print("Function {} took {} seconds to execute".format(func.__name__, end - start))
        return result
    return wrapper

def memoize(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if args not in wrapper.cache:
            wrapper.cache[args] = func(*args, **kwargs)
        return wrapper.cache[args]
    wrapper.cache = {}
    return wrapper

def count_calls(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        wrapper.num_calls += 1
        print("Call {} of {}".format(wrapper.num_calls, func.__name__))
        return func(*args, **kwargs)
    wrapper.num_calls = 0
    return wrapper



from typing import Any
from scanpy import AnnData
import torch
import gc
import pandas as pd 
import os 
from ..model import scAtlasVAE
from ..preprocessing._preprocess import annotate_t_cell_cd4_cd8
from ..utils._logger import mw, mt

class scAtlasVAEPipeline:
    def __init__(
        self,
        cd4cd8_model_checkpoint_path: str,
        cd8_model_checkpoint_path: str,
        cd4_model_checkpoint_path: str,
    ):
        """
        Pipeline for scAtlasVAE
        :param cd4cd8_model_checkpoint_path: str. Path to the checkpoint file of trained VAE model for CD4/CD8 datasets
        :param cd8_model_checkpoint_path: str. Path to the checkpoint file of trained VAE model for CD8 datasets
        :param cd4_model_checkpoint_path: str. Path to the checkpoint file of trained VAE model for CD4 datasets
        """
        if not os.path.exists(cd4cd8_model_checkpoint_path):
            raise ValueError(f"cd4cd8_model_checkpoint_path {cd4cd8_model_checkpoint_path} does not exist.")
        self.cd4cd8_model_checkpoint_path = cd4cd8_model_checkpoint_path
        if not os.path.exists(cd8_model_checkpoint_path):
            mw(f"cd8_model_checkpoint_path {cd8_model_checkpoint_path} does not exist.")
        self.cd8_model_checkpoint_path = cd8_model_checkpoint_path
        if not os.path.exists(cd4_model_checkpoint_path):
            mw(f"cd4_model_checkpoint_path {cd4_model_checkpoint_path} does not exist.")
        self.cd4_model_checkpoint_path = cd4_model_checkpoint_path

    def _single_thread_call(self, adata: AnnData, predict_label: bool = False, **kwargs: Any) -> AnnData:
        # CD4/CD8 model
        adata_cd4cd8 = scAtlasVAE.setup_anndata(adata, self.cd4cd8_model_checkpoint_path)
        state_dict = torch.load(self.cd4cd8_model_checkpoint_path, map_location="cpu")
        config = state_dict['model_config']
        config.update(kwargs)
        vae_model_cd4cd8 = scAtlasVAE(
            adata=adata_cd4cd8,
            pretrained_state_dict=state_dict['model_state_dict'],
            **config,
        )
        Z = vae_model_cd4cd8.get_latent_embedding()
        adata_cd4cd8.obsm["X_gex_CD4CD8"] = Z
        annotate_t_cell_cd4_cd8(adata_cd4cd8, use_rep="X_gex_CD4CD8")

        # CD4/CD8 Annotation
        adata_cd8 = adata_cd4cd8[adata_cd4cd8.obs["cd4_cd8_annotation"] == "CD8"]
        adata_cd4 = adata_cd4cd8[adata_cd4cd8.obs["cd4_cd8_annotation"] == "CD4"]
        
        # CD8 only model
        adata_cd8 = scAtlasVAE.setup_anndata(adata_cd8, self.cd8_model_checkpoint_path)
        state_dict = torch.load(self.cd8_model_checkpoint_path, map_location="cpu")
        config = state_dict['model_config']
        config.update(kwargs)
        vae_model_cd8 = scAtlasVAE(
            adata=adata_cd8,
            pretrained_state_dict=state_dict['model_state_dict'],
            **config,
        )
        Z = vae_model_cd8.get_latent_embedding()
        if (config['label_key'] is not None or config['additional_label_key'] is not None) and predict_label:
            predicted_labels = vae_model_cd8.predict_labels(return_pandas=True)
            predicted_labels.columns = list(map(lambda x: 'predicted_' + x, predicted_labels.columns))
            predicted_labels.index = adata_cd8.obs.index
            adata_cd8.obs = adata_cd8.obs.join(predicted_labels)
        adata_cd8.obsm["X_gex"] = Z

        # CD4 only model
        adata_cd4 = scAtlasVAE.setup_anndata(adata_cd4, self.cd4_model_checkpoint_path)
        state_dict = torch.load(self.cd4_model_checkpoint_path, map_location="cpu")
        config = state_dict['model_config']
        config.update(kwargs)
        vae_model_cd4 = scAtlasVAE(
            adata=adata_cd4,
            pretrained_state_dict=state_dict['model_state_dict'],
            **config,
        )
        Z = vae_model_cd4.get_latent_embedding()
        if (config['label_key'] is not None or config['additional_label_key'] is not None) and predict_label:
            predicted_labels = vae_model_cd4.predict_labels(return_pandas=True)
            predicted_labels.columns = list(map(lambda x: 'predicted_' + x, predicted_labels.columns))
            predicted_labels.index = adata_cd4.obs.index
            adata_cd4.obs = adata_cd4.obs.join(predicted_labels)
        adata_cd4.obsm["X_gex"] = Z
        
        return adata_cd4, adata_cd8
    
    def _multi_thread_call(self, adata: AnnData, predict_label: bool = False, n_per_block: int = 100000, **kwargs: Any) -> AnnData:
        # CD4/CD8 model
        _adata_cd4cd8 = []
        for i in range(0, adata.shape[0], n_per_block):
            adata_cd4cd8 = scAtlasVAE.setup_anndata(adata[i:i+n_per_block], self.cd4cd8_model_checkpoint_path)
            state_dict = torch.load(self.cd4cd8_model_checkpoint_path, map_location="cpu")
            config = state_dict['model_config']
            config.update(kwargs)
            vae_model_cd4cd8 = scAtlasVAE(
                adata=adata_cd4cd8,
                pretrained_state_dict=state_dict['model_state_dict'],
                **config,
            )
            Z = vae_model_cd4cd8.get_latent_embedding()
            adata_cd4cd8.obsm["X_gex_CD4CD8"] = Z
            _adata_cd4cd8.append(adata_cd4cd8)
            del vae_model_cd4cd8
            gc.collect()
            torch.cuda.empty_cache()

        adata_cd4cd8 = AnnData.concatenate(*_adata_cd4cd8)
        del _adata_cd4cd8
        gc.collect()
        torch.cuda.empty_cache()
        annotate_t_cell_cd4_cd8(adata_cd4cd8, use_rep="X_gex_CD4CD8")

        # CD4/CD8 Annotation
        adata_cd8_merged = adata_cd4cd8[adata_cd4cd8.obs["cd4_cd8_annotation"] == "CD8"]
        adata_cd4_merged = adata_cd4cd8[adata_cd4cd8.obs["cd4_cd8_annotation"] == "CD4"]
        
        # CD8 only model
        _adata_cd8 = []
        for i in range(0, adata_cd8_merged.shape[0], n_per_block):
            adata_cd8 = scAtlasVAE.setup_anndata(adata_cd8_merged[i:i+n_per_block], self.cd8_model_checkpoint_path)
            
            state_dict = torch.load(self.cd8_model_checkpoint_path, map_location="cpu")
            config = state_dict['model_config']
            config.update(kwargs)
            vae_model_cd8 = scAtlasVAE(
                adata=adata_cd8,
                pretrained_state_dict=state_dict['model_state_dict'],
                **config,
            )
            Z = vae_model_cd8.get_latent_embedding()
            if config['label_key'] is not None or config['additional_label_key'] is not None:
                predicted_labels = vae_model_cd8.predict_labels(return_pandas=True)
                predicted_labels.columns = list(map(lambda x: 'predicted_' + x, predicted_labels.columns))
                predicted_labels.index = adata_cd8.obs.index
            adata_cd8.obs = pd.concat([adata_cd8.obs, predicted_labels], 1)
            adata_cd8.obsm["X_gex"] = Z
            _adata_cd8.append(adata_cd8)
            del vae_model_cd8
            gc.collect()
            torch.cuda.empty_cache()

        adata_cd8 = AnnData.concatenate(*_adata_cd8)
        del _adata_cd8
        gc.collect()

        # CD4 only model
        _adata_cd4 = []
        for i in range(0, adata_cd4_merged.shape[0], n_per_block):
            adata_cd4 = scAtlasVAE.setup_anndata(adata_cd4_merged[i:i+n_per_block], self.cd4_model_checkpoint_path)
            state_dict = torch.load(self.cd4_model_checkpoint_path, map_location="cpu")
            config = state_dict['model_config']
            config.update(kwargs)
            vae_model_cd4 = scAtlasVAE(
                adata=adata_cd4,
                pretrained_state_dict=state_dict['model_state_dict'],
                **config,
            )
            Z = vae_model_cd4.get_latent_embedding()
            if config['label_key'] is not None or config['additional_label_key'] is not None:
                predicted_labels = vae_model_cd4.predict_labels(return_pandas=True)
                predicted_labels.columns = list(map(lambda x: 'predicted_' + x, predicted_labels.columns))
                predicted_labels.index = adata_cd4.obs.index
            adata_cd4.obs = pd.concat([adata_cd4.obs, predicted_labels], 1)
            adata_cd4.obsm["X_gex"] = Z
            _adata_cd4.append(adata_cd4)
            del vae_model_cd4
            gc.collect()
            torch.cuda.empty_cache()
        adata_cd4 = AnnData.concatenate(*_adata_cd4)
        del _adata_cd4
        gc.collect()
        torch.cuda.empty_cache()
        
        return adata_cd4, adata_cd8


    def __call__(self, adata: AnnData, predict_label: bool = False, **kwargs: Any ) -> AnnData:
        if adata.shape[0] <= 100000:
            return self._single_thread_call(adata, predict_label, **kwargs)
        else:
            return self._multi_thread_call(adata, predict_label, **kwargs)


import setuptools 


version = '0.0.8'
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
    
setuptools.setup(
    name="scAtlasVAE",
    version=version,
    author="Ziwei Xue",
    author_email="xueziweisz@gmail.com",
    description="scAtlasVAE: a deep learning framework for atlas-scale scRNA-seq integration and analysis",
    long_description="scAtlasVAE: a deep learning framework for atlas-scale scRNA-seq integration and analysis",
    packages=setuptools.find_packages(exclude=[
        "*reference*",
        "*pretrained_weights*",
        "*docs*",
    ]),
    install_requires=[
        'anndata==0.8.0',
        'scanpy==1.8.1',
        'scikit-learn==0.24.1',
        'matplotlib==3.3.4',
        'einops==0.4.1 ',
        'biopython==1.79',
        'seaborn==0.12.2',
        'scirpy==0.10.1',
        'pandas==1.4.2',
        'tabulate==0.8.9',
        'umap-learn',
    ],
    dependency_links=['https://download.pytorch.org/whl/cu117'],
    include_package_data=False,
)



import os
from multiprocessing import Manager
import warnings
import scanpy as sc
from threading import Thread
from typing import (
    Any, Callable, Optional, Sequence, Union, Iterable
)
from joblib import delayed, Parallel
import numpy as np
from scipy.sparse import issparse
import pandas as pd 

class Parallelizer:
    def __init__(self, n_jobs:int):
        self.n_jobs = self.get_n_jobs(n_jobs=n_jobs)
        self._msg_shown = False 
        
    def get_n_jobs(self, n_jobs):
        if n_jobs is None or (n_jobs < 0 and os.cpu_count() + 1 + n_jobs <= 0):
            return 1
        elif n_jobs > os.cpu_count():
            return os.cpu_count()
        elif n_jobs < 0:
            return os.cpu_count() + 1 + n_jobs
        else:
            return n_jobs

    @staticmethod
    def __update__(pbar, queue, n_total):
        n_finished = 0
        while n_finished < n_total:
            try:
                res = queue.get()
            except EOFError as e:
                if not n_finished != n_total:
                    raise RuntimeError(
                        f"Finished only `{n_finished} out of `{n_total}` tasks.`"
                    ) from e
                break
            assert res in (None, (1, None), 1)  # (None, 1) means only 1 job
            if res == (1, None):
                n_finished += 1
                if pbar is not None:
                    pbar.update()
            elif res is None:
                n_finished += 1
            elif pbar is not None:
                pbar.update()

        if pbar is not None:
            pbar.close()
        
    def parallelize(
        self,
        map_func: Callable[[Any], Any],
        map_data: Union[Sequence[Any], Iterable[Any]],
        n_split: Optional[int] = None,
        progress: bool = True,
        progress_unit: str = "",
        use_ixs: bool = False,
        backend: str = "loky",
        reduce_func: Optional[Callable[[Any], Any]] = None,
        reduce_as_array: bool = True,
    ):
        if progress:
            try:
                try:
                    from tqdm.notebook import tqdm
                except ImportError:
                    try:
                        from tqdm import tqdm_notebook as tqdm
                    except ImportError:
                        from tqdm import tqdm
                import ipywidgets  # noqa
            except ImportError:
                global _msg_shown
                tqdm = None

                self._msg_shown = True
            else:
                tqdm = None

        col_len = map_data.shape[0] if issparse(map_data) else len(map_data)

        if n_split is None:
            n_split = self.n_jobs

        if issparse(map_data):
            if n_split == map_data.shape[0]:
                map_datas = [map_data[[ix], :] for ix in range(map_data.shape[0])]
            else:
                step = map_data.shape[0] // n_split

                ixs = [
                    np.arange(i * step, min((i + 1) * step, map_data.shape[0]))
                    for i in range(n_split)
                ]
                ixs[-1] = np.append(
                    ixs[-1], np.arange(ixs[-1][-1] + 1, map_data.shape[0])
                )

                map_datas = [map_data[ix, :] for ix in filter(len, ixs)]
        else:
            map_datas = list(filter(len, np.array_split(map_data, n_split)))

        pass_queue = not hasattr(map_func, "py_func")  # we'd be inside a numba function

        def wrapper(*args, **kwargs):
            if pass_queue and progress:
                pbar = None if tqdm is None else tqdm(total=col_len, unit=progress_unit)
                queue = Manager().Queue()
                thread = Thread(target=Parallelizer.__update__, args=(pbar, queue, len(map_datas)))
                thread.start()
            else:
                pbar, queue, thread = None, None, None

            res = Parallel(n_jobs=self.n_jobs, backend=backend)(
                delayed(map_func)(
                    *((i, cs) if use_ixs else (cs,)),
                    *args,
                    **kwargs,
                    queue=queue,
                )
                for i, cs in enumerate(map_datas)
            )

            res = np.array(res) if reduce_as_array else res
            if thread is not None:
                thread.join()

            return res if reduce_func is None else reduce_func(res)
        return wrapper
    
def parallel_leiden_computation(
    X: np.ndarray,
    n_neighbors_list: Iterable[int] = (30, 50, 70, 90),
    resolution_list: Iterable[float] = (0.4, 0.6, 0.8, 1.0, 1.2),
):
    def par_func(data, queue=None):
        out = {}
        with warnings.catch_warnings():
            warnings.simplefilter("ignore") 
            for i in data:
                X, n_neighbors, resolution = i["X"], i["n_neighbors"], i["resolution"]
                adata = sc.AnnData(X=X)
                sc.pp.neighbors(adata, n_neighbors=n_neighbors)
                sc.tl.leiden(adata, resolution=resolution)
                if queue is not None:
                    queue.put(None)
                out[(n_neighbors, resolution)] = adata.obs["leiden"].values
        return out 
    map_data=[{"X": X, "n_neighbors": n_neighbors, "resolution": resolution} for n_neighbors in n_neighbors_list for resolution in resolution_list]
    p = Parallelizer(n_jobs=min(len(map_data), os.cpu_count()))
    result = p.parallelize(
        map_func=par_func, 
        map_data=map_data, 
        reduce_func=lambda x: x
    )()
    return pd.DataFrame(
        list(map(lambda x: list(x.values())[0], result)), 
        columns=list(map(lambda x: list(x.keys())[0], result))
    )




import scanpy as sc
import infercnvpy as cnv
import pandas as pd
from typing import Callable, List, Optional, Union, Iterable
from pathlib import Path
import json
import os
import numpy as np
import uuid
import tqdm
from ..utils._logger import mt
from ..utils._compat import Literal
from ..utils._utilities import writeMM

MODULE_PATH = Path(__file__).parent

HSAP_GENE_ORDER = MODULE_PATH / '../data/refdata/human/hsap_gene_order.txt'



def infercnv_R(
    adata:sc.AnnData,
    obs_key:str,
    ref_group_names: Optional[List[str]] = None,
    platform: Literal['10x', 'smartseq2'] = '10x',
    r_path: str = "/Users/snow/.conda/envs/r413/bin/Rscript", 
):
    if len(np.unique(adata.obs.index)) != adata.shape[0]:
        adata.obs_names_make_unique()
    tempfile = '/tmp/infercnv/' + str(uuid.uuid4())
    if not os.path.exists(tempfile):
        os.makedirs(tempfile)
    X = adata.X
    mt(f"Write raw_counts_matrix.mat and annotations_file.tsv to {tempfile}")
    writeMM(X, tempfile + '/raw_counts_matrix.mat')

    pd.DataFrame(adata.obs.index).to_csv(
        tempfile + '/obsnames.tsv',
        sep='\t',
        header=None,
        index=False
    )
    pd.DataFrame(adata.var.index).to_csv(
        tempfile + '/varnames.tsv',
        sep='\t',
        header=None,
        index=False
    )
    mt(f"Write annotations_file.tsv to {tempfile}")
    adata.obs.loc[:,obs_key].to_csv(
        tempfile + '/annotations_file.tsv',
        sep='\t',
        header=None
    )
    r_script = f"""
library(infercnv)
raw_counts_matrix <- Matrix::readMM("{tempfile + '/raw_counts_matrix.mat'}")
raw_counts_matrix <- as(raw_counts_matrix, "dgCMatrix")
rownames(raw_counts_matrix) <- read.csv("{tempfile + '/varnames.tsv'}",header=FALSE)[,1]
colnames(raw_counts_matrix) <- read.csv("{tempfile + '/obsnames.tsv'}",header=FALSE)[,1]
infercnv_obj = CreateInfercnvObject(
    # raw_counts_matrix="{tempfile + '/raw_counts_matrix.tsv.gz'}",
    raw_counts_matrix=raw_counts_matrix,
    annotations_file="{tempfile + '/annotations_file.tsv'}",
    delim="\t",
    gene_order_file="{HSAP_GENE_ORDER}",
    ref_group_names = c({json.dumps(ref_group_names)[1:-1]})) 

infercnv_obj = infercnv::run(
    infercnv_obj,
    cutoff={1 if platform == '10x' else 0.1},
    out_dir="{tempfile}", 
    cluster_by_groups=TRUE, 
)
    """
    mt(f"Write rscript to {tempfile}")
    with open(os.path.join(tempfile, "rscript.R"), "w+") as f:
        f.write(r_script)
    mt(f"Run rscript")
    os.system(f'{r_path} {os.path.join(tempfile, "rscript.R")}')


def infercnv_py(
    adata:sc.AnnData,
    obs_key:str,
    ref_group_names: Optional[List[str]] = None,
    platform: Literal['10x', 'smartseq2'] = '10x',
    window_size: int = 250,
    n_jobs: int = 1,
):
    gene_order = pd.read_csv(HSAP_GENE_ORDER, sep='\t', header=None, index_col=0)
    gene_order.columns = ["chromosome", "start", "end"]
    adata = adata[:, list(filter(lambda x: x in gene_order.index, adata.var.index))]
    adata.var = adata.var.join(gene_order, how='left')
    cnv.tl.infercnv(
        adata,
        reference_key=obs_key,
        reference_cat=ref_group_names,
        window_size=window_size,
        n_jobs=n_jobs,
    )
    return adata

#scores=apply(infercnv_obj@expr.data,2,function(x){ sum(x < 0.95 | x > 1.05)/length(x) })


import torch
from torch.distributions import constraints, NegativeBinomial, Poisson, Beta, Normal
from torch.distributions.utils import (
    broadcast_all,
    lazy_property,
    logits_to_probs,
    probs_to_logits,
)
from torch.nn.functional import softplus
from torch.distributions.distribution import Distribution
from scipy.stats import f
import numpy as np

def multivariate_gaussian_p_value(X, Y):
    # Step 1: Compute the mean vectors for X and Y
    mean_X = np.mean(X, axis=0)
    mean_Y = np.mean(Y, axis=0)

    # Step 2: Compute the covariance matrices for X and Y (use unbiased estimator)
    cov_X = np.cov(X, rowvar=False, bias=False)
    cov_Y = np.cov(Y, rowvar=False, bias=False)

    # Step 3: Compute the sample sizes of X and Y
    n_X = X.shape[0]
    n_Y = Y.shape[0]

    # Step 4: Compute the Hotelling's T-square test statistic
    t_square = (n_X * n_Y) / (n_X + n_Y) * np.dot(np.dot((mean_X - mean_Y), np.linalg.inv((cov_X / n_X) + (cov_Y / n_Y))), (mean_X - mean_Y))

    # Step 5: Calculate the degrees of freedom for the F-distribution
    df1 = X.shape[1]  # Number of dimensions
    df2 = n_X + n_Y - df1 - 1

    # Step 6: Calculate the p-value using the cumulative distribution function (CDF) of the F-distribution
    p_value = 1.0 - f.cdf(t_square, df1, df2)

    return p_value


# Pyro Distributions.utils
def broadcast_shape(*shapes, **kwargs):
    """
    Similar to ``np.broadcast()`` but for shapes.
    Equivalent to ``np.broadcast(*map(np.empty, shapes)).shape``.
    :param tuple shapes: shapes of tensors.
    :param bool strict: whether to use extend-but-not-resize broadcasting.
    :returns: broadcasted shape
    :rtype: tuple
    :raises: ValueError
    """
    strict = kwargs.pop("strict", False)
    reversed_shape = []
    for shape in shapes:
        for i, size in enumerate(reversed(shape)):
            if i >= len(reversed_shape):
                reversed_shape.append(size)
            elif reversed_shape[i] == 1 and not strict:
                reversed_shape[i] = size
            elif reversed_shape[i] != size and (size != 1 or strict):
                raise ValueError(
                    "shape mismatch: objects cannot be broadcast to a single shape: {}".format(
                        " vs ".join(map(str, shapes))
                    )
                )
    return tuple(reversed(reversed_shape))

# Pyro Distributions.zero_inflated
class ZeroInflatedDistribution(Distribution):
    """
    Generic Zero Inflated distribution.
    This can be used directly or can be used as a base class as e.g. for
    :class:`ZeroInflatedPoisson` and :class:`ZeroInflatedNegativeBinomial`.
    :param TorchDistribution base_dist: the base distribution.
    :param torch.Tensor gate: probability of extra zeros given via a Bernoulli distribution.
    :param torch.Tensor gate_logits: logits of extra zeros given via a Bernoulli distribution.
    """

    arg_constraints = {
        "gate": constraints.unit_interval,
        "gate_logits": constraints.real,
    }

    def __init__(self, base_dist, *, gate=None, gate_logits=None, validate_args=None):
        if (gate is None) == (gate_logits is None):
            raise ValueError(
                "Either `gate` or `gate_logits` must be specified, but not both."
            )
        if gate is not None:
            batch_shape = broadcast_shape(gate.shape, base_dist.batch_shape)
            self.gate = gate.expand(batch_shape)
        else:
            batch_shape = broadcast_shape(gate_logits.shape, base_dist.batch_shape)
            self.gate_logits = gate_logits.expand(batch_shape)
        if base_dist.event_shape:
            raise ValueError(
                "ZeroInflatedDistribution expected empty "
                "base_dist.event_shape but got {}".format(base_dist.event_shape)
            )

        self.base_dist = base_dist.expand(batch_shape)
        event_shape = torch.Size()

        super().__init__(batch_shape, event_shape, validate_args)

    @constraints.dependent_property
    def support(self):
        return self.base_dist.support

    @lazy_property
    def gate(self):
        return logits_to_probs(self.gate_logits)

    @lazy_property
    def gate_logits(self):
        return probs_to_logits(self.gate)

    def log_prob(self, value):
        if self._validate_args:
            self._validate_sample(value)

        if "gate" in self.__dict__:
            gate, value = broadcast_all(self.gate, value)
            log_prob = (-gate).log1p() + self.base_dist.log_prob(value)
            log_prob = torch.where(value == 0, (gate + log_prob.exp()).log(), log_prob)
        else:
            gate_logits, value = broadcast_all(self.gate_logits, value)
            base_log_prob = self.base_dist.log_prob(value)
            log_prob_minus_log_gate = -gate_logits + base_log_prob
            log_gate = -softplus(-gate_logits)
            log_prob = log_prob_minus_log_gate + log_gate
            
            zero_log_prob = softplus(log_prob_minus_log_gate) + log_gate
            log_prob = torch.where(value == 0, zero_log_prob, log_prob)
        return log_prob

    def sample(self, sample_shape=torch.Size()):
        shape = self._extended_shape(sample_shape)
        with torch.no_grad():
            mask = torch.bernoulli(self.gate.expand(shape)).bool()
            samples = self.base_dist.expand(shape).sample()
            samples = torch.where(mask, samples.new_zeros(()), samples)
        return samples

    @lazy_property
    def mean(self):
        return (1 - self.gate) * self.base_dist.mean

    @lazy_property
    def variance(self):
        return (1 - self.gate) * (
            self.base_dist.mean ** 2 + self.base_dist.variance
        ) - (self.mean) ** 2

    def expand(self, batch_shape, _instance=None):
        new = self._get_checked_instance(type(self), _instance)
        batch_shape = torch.Size(batch_shape)
        gate = self.gate.expand(batch_shape) if "gate" in self.__dict__ else None
        gate_logits = (
            self.gate_logits.expand(batch_shape)
            if "gate_logits" in self.__dict__
            else None
        )
        base_dist = self.base_dist.expand(batch_shape)
        ZeroInflatedDistribution.__init__(
            new, base_dist, gate=gate, gate_logits=gate_logits, validate_args=False
        )
        new._validate_args = self._validate_args
        return new

# Pyro Distributions.zero_inflated
class ZeroInflatedPoisson(ZeroInflatedDistribution):
    """
    A Zero Inflated Poisson distribution.
    :param torch.Tensor rate: rate of poisson distribution.
    :param torch.Tensor gate: probability of extra zeros.
    :param torch.Tensor gate_logits: logits of extra zeros.
    """

    arg_constraints = {
        "rate": constraints.positive,
        "gate": constraints.unit_interval,
        "gate_logits": constraints.real,
    }
    support = constraints.nonnegative_integer

    def __init__(self, rate, *, gate=None, gate_logits=None, validate_args=None):
        base_dist = Poisson(rate=rate, validate_args=False)
        base_dist._validate_args = validate_args

        super().__init__(
            base_dist, gate=gate, gate_logits=gate_logits, validate_args=validate_args
        )

    @property
    def rate(self):
        return self.base_dist.rate

# Pyro Distributions.zero_inflated
class ZeroInflatedNegativeBinomial(ZeroInflatedDistribution):
    """
    A Zero Inflated Negative Binomial distribution.
    :param total_count: non-negative number of negative Bernoulli trials.
    :type total_count: float or torch.Tensor
    :param torch.Tensor probs: Event probabilities of success in the closed interval [0, 1].
    :param torch.Tensor logits: Event log-odds for probabilities of success.
    :param torch.Tensor gate: probability of extra zeros.
    :param torch.Tensor gate_logits: logits of extra zeros.
    """

    arg_constraints = {
        "total_count": constraints.greater_than_eq(0), # Should the prob be half-open?
        "probs": constraints.interval(0.0, 1.0),
        "logits": constraints.real,
        "gate": constraints.unit_interval,
        "gate_logits": constraints.real,
    }
    support = constraints.nonnegative_integer

    def __init__(
        self,
        total_count,
        *,
        probs=None,
        logits=None,
        gate=None,
        gate_logits=None,
        validate_args=None
    ):

        base_dist = NegativeBinomial(
            total_count=total_count,
            probs=probs,
            logits=logits,
            validate_args=False,
        )
        base_dist._validate_args = validate_args

        super().__init__(
            base_dist, gate=gate, gate_logits=gate_logits, validate_args=validate_args
        )

    @property
    def total_count(self):
        return self.base_dist.total_count

    @property
    def probs(self):
        return self.base_dist.probs

    @property
    def logits(self):
        return self.base_dist.logits

class ZeroInflatedGaussian(ZeroInflatedDistribution):
    arg_constraints = {
        "mean": constraints.real,
        "variance": constraints.greater_than(0),
        "gate": constraints.unit_interval,
        "gate_logits": constraints.real,
    }

    def __init__(
        self,
        mean,
        *,
        variance,
        gate=None,
        gate_logits=None,
        validate_args=None
    ) -> None:
        base_dist = Normal(mean, variance)
        base_dist._validate_args = validate_args
        super().__init__(
            base_dist, gate=gate, gate_logits=gate_logits, validate_args=validate_args
        )
    
    @property
    def mean(self):
        return self.base_dist.mean 
    
    @property
    def variance(self):
        return self.base_dist.variance

class ZeroInflatedBeta(ZeroInflatedDistribution):
    arg_constraints = {
        "concentration1": constraints.real,
        "concentration2": constraints.real,
        "gate": constraints.unit_interval,
        "gate_logits": constraints.real
    }

    def __init__(
        self,
        concentration1,
        concentration2,
        gate=None,
        gate_logits=None,
        validate_args=None
    ) -> None:  
        base_dist = Beta(concentration1, concentration2)
        super().__init__(
            base_dist, gate=gate, gate_logits=gate_logits, validate_args=validate_args
        )
    
    @property
    def mean(self):
        return self.concentration1 / (self.concentration1 + self.concentration2)
    
    @property
    def variance(self):
        return (self.concentration1 * self.concentration2) / ((self.concentration1 + self.concentration2) ** 2 + (self.concentration1 + self.concentration2 + 1))


from enum import Enum

class FCDEF(Enum):
    ENCODER = 0x0
    DECODER = 0x1

TRA_DEFINITION = ["CDR3a","TRAV","TRAJ"]
TRB_DEFINITION = ["CDR3b","TRBV","TRBJ"]
TRAB_DEFINITION = ["CDR3a","CDR3b","TRAV","TRAJ","TRBV","TRBJ"]


TRA_DEFINITION_ORIG = ["IR_VJ_1_junction_aa",
    "IR_VJ_1_v_call",
    "IR_VJ_1_j_call"
]
TRB_DEFINITION_ORIG = [
    "IR_VDJ_1_junction_aa",
    "IR_VDJ_1_v_call",
    "IR_VDJ_1_j_call"
]
TRAB_DEFINITION_ORIG = [
    "IR_VJ_1_junction_aa",            
    "IR_VDJ_1_junction_aa",
    "IR_VJ_1_v_call",
    "IR_VJ_1_j_call",
    "IR_VDJ_1_v_call",
    "IR_VDJ_1_j_call"
]

CDR1a = [
    'TSGFYG',
    'TSGFNG',
    'VSNAYN',
    'VSGNPY',
    'NIATNDY',
    'DSSSTY',
    'NYSPAY',
    'VSRFNN',
    'YGGTVN',
    'SSYSPS',
    'YGATPY',
    'SSVPPY',
    'SSVSVY',
    'YSGVPS',
    'TTQYPS',
    'ATGYPS',
    'VSPFSN',
    'ERTLFN',
    'NSASQS',
    'DRGSQS',
    'NSAFQY',
    'DSASNY',
    'NSASDY',
    'TSDPSYG',
    'YSGSPE',
    'TSINN',
    'SSYSTF',
    'TRDTTYY',
    'VSGLRG',
    'DSAIYN',
    'DSVNN',
    'NTAFDY',
    'SSNFYA',
    'TTLSN',
    'TISGNEY',
    'TISGTDY',
    'SVFSS',
    'NSMFDY',
    'KALYS',
    'KTLYG',
    'SIFNT',
    'VTNFRS',
    'TSENNYY',
    'TSESDYY',
    'TTSDR',
    'STGYPT',
    'VGISA'
]

CDR2a = ['NALDGL',
'NVLDGL',
'GSKP',
'YITGDNLV',
'GYKTK',
'IFSNMDM',
'IRENEKE',
'MYSAGYE',
'YFSGDPLV',
'YTSAATLV',
'YFSGDTLV',
'YTSAATLV',
'YLSGSTLV',
'DLTEATQV',
'AMKANDK',
'ATKADDK',
'MTFSENT',
'IQSSQKE',
'VYSSGN',
'IYSNGD',
'TYSSGN',
'IRSNVGE',
'IRSNMDK',
'QGSYDQQN',
'HISR',
'IRSNERE',
'SSENQE',
'RNSFDEQN',
'LYSAGEE',
'IQSSQRE',
'IPSGT',
'IRPDVSE',
'MTLNGDE',
'LVKSGEV',
'GLKNN',
'GLTSN',
'VVTGGEV',
'ISSIKDK',
'LLKGGEQ',
'LQKGGEE',
'LYKAGEL',
'LTSSGIE',
'QEAYKQQN',
'QEAYKQQN',
'LLSNGAV',
'ETME',
'LSSGK']

CDR1b = ['GHDS',
'SNHLY',
'LGHDT',
'LGHNA',
'MGHRA',
'LGHNA',
'LGHNA',
'SGHRS',
'SGHSS',
'SGHNT',
'SGHKS',
'SGHDT',
'SGHTS',
'SGHTS',
'MNHNS',
'MNHEY',
'MNHEY',
'MRHNA',
'MNHEY',
'MNHNY',
'MNHEY',
'MNHGY',
'MNHGY',
'SGHNA',
'SGHTA',
'SGHTA',
'SGHVT',
'SGHVS',
'SSHAT',
'SGHVS',
'SEHNR',
'SGDLS',
'WNHNN',
'WSHSY',
'ENHRY',
'SGHAT',
'SGHAT',
'SGHNT',
'SGHND',
'FGHNF',
'SGHNS',
'SGHDY',
'LGHNT',
'PRHDT',
'SGHDN',
'LNHNV',
'KGHSY',
'SGHMF',
'KGHSH',
'LNHDA',
'DFQATT',
'KAHSY',
'KGHTF',
'KGHDR',
'MGHDK',
'MNHVT',
'MNHEY',
'MDHEN',
'SQVTM',
'GTSNPN',
'FQATT']

CDR2b = ['YNCKEF',
'FYNNEI',
'YNNKEL',
'YSNKEP',
'YSYEKL',
'YNFKEQ',
'YSLEER',
'YFSETQ',
'YANELR',
'YYREEE',
'YYEKEE',
'YYEEEE',
'YYEKEE',
'YDEGEE',
'SASEGT',
'SVGEGT',
'SVGEGT',
'SNTAGT',
'SVGAGI',
'SVGAGI',
'SVAAAL',
'SAAAGT',
'SVAAGI',
'FQGKDA',
'FQGNSA',
'FQGTGA',
'SQSDAQ',
'FNYEAQ',
'FNYEAQ',
'FQNEAQ',
'FQNEAQ',
'YYNGEE',
'SYGVQD',
'SAAADI',
'SYGVKD',
'FQDESV',
'FQNNGV',
'YENEEA',
'FCSWTL',
'FRS*SI',
'FNNNVP',
'FNNNVP',
'FRNRAP',
'FYEKMQ',
'FVKESK',
'YYDKDF',
'FQNENV',
'FQYQNI',
'LQKENI',
'SQIVND',
'SNEGSKA',
'FQNEEL',
'FQNEQV',
'SFDVKD',
'SYGVNS',
'SPGTGS',
'SMNVEV',
'SYDVKM',
'ANQGSEA',
'SVGIG',
'SNEGSEV']

TRAV_KEYS = ['TRAV1-1',
'TRAV1-2',
'TRAV2',
'TRAV3',
'TRAV4',
'TRAV5',
'TRAV6',
'TRAV7',
'TRAV8-1',
'TRAV8-2',
'TRAV8-3',
'TRAV8-4',
'TRAV8-6',
'TRAV8-7',
'TRAV9-1',
'TRAV9-2',
'TRAV10',
'TRAV11',
'TRAV12-1',
'TRAV12-2',
'TRAV12-3',
'TRAV13-1',
'TRAV13-2',
'TRAV14DV4',
'TRAV16',
'TRAV17',
'TRAV18',
'TRAV19',
'TRAV20',
'TRAV21',
'TRAV22',
'TRAV23DV6',
'TRAV24',
'TRAV25',
'TRAV26-1',
'TRAV26-2',
'TRAV27',
'TRAV29DV5',
'TRAV30',
'TRAV34',
'TRAV35',
'TRAV36DV7',
'TRAV38-1',
'TRAV38-2DV8',
'TRAV39',
'TRAV40',
'TRAV41',]

TRBV_KEYS = ['TRBV1',
'TRBV2',
'TRBV3-1',
'TRBV3-2',
'TRBV4-1',
'TRBV4-2',
'TRBV4-3',
'TRBV5-1',
'TRBV5-3',
'TRBV5-4',
'TRBV5-5',
'TRBV5-6',
'TRBV5-7',
'TRBV5-8',
'TRBV6-1',
'TRBV6-2',
'TRBV6-3',
'TRBV6-4',
'TRBV6-5',
'TRBV6-6',
'TRBV6-7',
'TRBV6-8',
'TRBV6-9',
'TRBV7-1',
'TRBV7-2',
'TRBV7-3',
'TRBV7-4',
'TRBV7-6',
'TRBV7-7',
'TRBV7-8',
'TRBV7-9',
'TRBV9',
'TRBV10-1',
'TRBV10-2',
'TRBV10-3',
'TRBV11-1',
'TRBV11-2',
'TRBV11-3',
'TRBV12-1',
'TRBV12-2',
'TRBV12-3',
'TRBV12-4',
'TRBV12-5',
'TRBV13',
'TRBV14',
'TRBV15',
'TRBV16',
'TRBV17',
'TRBV18',
'TRBV19',
'TRBV20-1',
'TRBV21-1',
'TRBV23-1',
'TRBV24-1',
'TRBV25-1',
'TRBV26',
'TRBV27',
'TRBV28',
'TRBV29-1',
'TRBV30',
'TRBV29OR9-2']

TRAV2CDR1a = dict(zip(TRAV_KEYS, CDR1a))
TRAV2CDR2a = dict(zip(TRAV_KEYS, CDR2a))
TRBV2CDR1b = dict(zip(TRBV_KEYS, CDR1b))
TRBV2CDR2b = dict(zip(TRBV_KEYS, CDR2b))
